/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
   Module name:    window.c

   This module handles windows: creating, destroying, printing on windows.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <curses.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/stat.h>
#include <sys/ioctl.h>

#include "../config.h"

#include "../common/cmalloc.h"
#include "../common/common.h"
#include "../common/fonts.h"

#include "../siag/calc.h"
#include "../siag/window.h"
#include "../siod/siod.h"

#include "tsiag.h"

#ifdef GUILE
#include <guile/gh.h>
#endif

#ifdef TCL
#include <tcl.h>
#endif

#define KB_MAX 10000
#define IQ_MAX 10000

struct ncurses_info {
	WINDOW *win;
	struct winsize size;
} single;

int input_warp_pointer = 1;	/* move cursor to input field */

int macro_flag;

textbuf kbd_macro = {0, 0, 0};

struct {
	char *buf;
	int index, size;
} input_queue;

window *w_list;

sfmt fmt0 = {"Helvetica", 100, 0, 0, 0, "black", "white", 0, 0, 0, 0};
sfmt fmt1 = {"Helvetica", 120, 0, 0, 0, "black", "grey", 0, VADJ_CENTER, HADJ_CENTER, 0};

int pr_scr_flag;
int grid_lines = 1;

int recalc;	/* number of iterations */

int columns;    /* number of cells per line */
int width;      /* default cell width */
int left_margin;
/*int right_margin = 5;*/

int oldr, oldc;

int rowcol_fmt_index;

unsigned int rowcol_format()
{
	return rowcol_fmt_index;        /* set in init_windows */
}

/* Prints line i from the top, starting at 0. */
void pr_line(window *w, int i)
{
	int j;

	wstandout(single.win);
	wmove(single.win, 1 + i, 0);
	wprintw(single.win, "%-*d", left_margin - 1, get_top(w).row + i);

	for (j = 0; j < columns; j++) {
		char s[1024];
		ret_pvalue(s, buffer_of_window(w), w->sht,
		           get_top(w).row + i,
		           get_top(w).col + j, -1);
		if (inblock(w, get_top(w).row + i, get_top(w).col + j))
			wstandout(single.win);
		else
			wstandend(single.win);
		wmove(single.win, 1 + i, left_margin + width * j);
		waddnstr(single.win, s, single.size.ws_col - left_margin -
		                        (width * j));
	}

	wstandend(single.win);
} /* pr_line */

/* Print and refresh all the windows. Sets pr_scr_flag to FALSE. */
void pr_scr(window *w)
{
	buffer *b;
	b = b_list;

	TRACEME((f,"pr_scr(): b->recalc=%d, recalc=%d", b->recalc, recalc));

	do {
		if (b->recalc) {
			int i;
			b->recalc = 0;
			for (i = 0; i < recalc; i++) calc_matrix(b);
		}
		b = b->next;
	} while (b != b_list);

	werase(single.win);
  
	do {
		int i;

		wstandout(single.win);

		for (i = 0; i < columns; i++) {
			wmove(single.win, 0, left_margin + width * i);
			wprintw(single.win, "%-*s", width - 1,
			        colnum_text(get_top(w).col + i));
		}
		for (i = 0; i < single.size.ws_row - 2; i++) pr_line(w, i);

		wstandend(single.win);

	} while ((w = next_window(w)) != w_list);

	pr_scr_flag = FALSE;
} /* pr_scr */

/*X
Print prompt on the bottom line and let the user enter a line into buffer.
The following commands exist:
a .. ~		Insert a character
<tab>		Call comp(buffer)
C-t		Interchange the characters around cursor
C-b		Move one character backward
C-f		Move one character forward
C-a		Move to the beginning of the line
C-e		Move to the end of the line
C-d		Delete the character at the cursor
C-h		Delete the character before the cursor
C-k		Delete the rest of the line
Escape		Delete the whole line
C-g		Print "Quit" and return FALSE
Enter	 	Return TRUE
X*/
int ask_for_str_comp(char *prompt, char *buffr, int (*comp)(char *))
{
	int l = strlen(buffr);

	TRACEME((f, "ask_for_str_comp: %s", prompt));

	for (;;) {
		int c;

		wmove(single.win, single.size.ws_row - 1, 0);
		waddstr(single.win, prompt);
		wclrtoeol(single.win);

		/* Show the buffer */
		if (l + strlen(prompt) < single.size.ws_col) {
			waddnstr(single.win, buffr, single.size.ws_col - strlen(prompt));
		} else {
			waddstr(single.win, &buffr[l - single.size.ws_col + strlen(prompt) + 1]);
		}

		/* Then place the cursor */
		if (l + strlen(prompt) < single.size.ws_col) {
			wmove(single.win, single.size.ws_row - 1, strlen(prompt) + l);
		} else {
			wmove(single.win, single.size.ws_row - 1, single.size.ws_col - 1);
		}

		wrefresh(single.win);

		switch (c = get_char_from_input_queue()) {
		case CTRL('a'): /* Move to the beginning of the line */
			l = 0; break;
		case CTRL('e'): /* Move to the end of the line */
			l = strlen(buffr); break;
		case CTRL('b'): /* Move one character backward */
			if (l > 0) l--; break;
		case CTRL('f'): /* Move one character forward */
			if (buffr[l]) l++; break;
		case CTRL('k'): /* Delete the rest of the line */
			buffr[l] = '\0'; break;
		case CTRL('d'): /* Delete the character at cursor */
			if (buffr[l]) {
				/* strcpy(buffr+l, buffr+l+1); */
				memmove(buffr+l,buffr+l+1,strlen(buffr+l+1)+1);
			}
			break;
		case CTRL('h'): /* Delete the character before cursor */
			if (l > 0) {
				l--;
				/* strcpy(buffr+l, buffr+l+1); */
				memmove(buffr+l,buffr+l+1,strlen(buffr+l+1)+1);
			}
			break;
		case CTRL('g'): /* Return and don't save changes */
			llpr("Changes abandoned"); return FALSE;
		case '\t': /* Call comp(buffer) */
			if ((*comp)(buffr)) return TRUE;
			l = strlen(buffr);
			break;
		case CTRL('t'): /* Interchange the characters around cursor */
			/* umm... I don't see why this is useful. */
			/* I'm leaving it here, though :) */
			if (strlen(buffr) >= 2) {
				if (buffr[l]) l++;
				if (l == 1) {
					c = buffr[0];
					buffr[0] = buffr[1];
					buffr[1] = c;
				} else {
					c = buffr[l-1];
					buffr[l-1] = buffr[l-2];
					buffr[l-2] = c;
				}
			}
			break;
		case ESC: buffr[0] = '\0'; break;
		case '\n': return TRUE; /* Accept changes */
		default: /* Normal key press */
			if (c >= ' ' && c <= '~' ) {
				int i;
				for (i = strlen(buffr) + 1; i > l; i--)
					buffr[i] = buffr[i - 1];
				buffr[l++] = c;
#ifdef HAVE_BEEP
			} else {
				beep(); /* hehehe */
#endif
			}
			break;
		}
	}
	/*NOTREACHED*/
}

/* EXTERNAL */
int edit_cell(char *p, char *b) {
	oldr = oldc = -1; /* Make show_cur redraw the contents */
	return ask_for_str(p, b);
}

/* Adds c to the keyboard macro.
Macros longer than KB_MAX charactesr are not allowed. */
void record_macro(int c)
{
	if (kbd_macro.size >= KB_MAX) {
		llpr("Macro too long; definition terminated");
		kbd_macro.size = 0;
		macro_flag = FALSE;
		return;
	}

	kbd_macro.text[kbd_macro.size++] = c;
}

/*X
Writes the buffer to the input queue.  The size of the buffer is limited
to IQ_MAX characters to prevent recursive macros from crashing the program
(the macros crash, of course, but that is less of a disaster).
TRUE is returned if the string could be added, FALSE otherwise.
X*/
int add_str_to_input_queue(textbuf buf)
{
	int i, j;

	if (input_queue.size + buf.size >= IQ_MAX) return FALSE;

	i = 0;
	j = input_queue.index + input_queue.size;

	while (i < buf.size) {
		/* this isn't very fast, but it should be fast */
		/* enough for our needs */

		j %= IQ_MAX;	/* wrap to beginning of text */
		input_queue.buf[j++] = buf.text[i++];
		input_queue.size++;
	}

	return TRUE;
}

/* This where we fix input, so that things like arrow keys work. */
int fix_input(int c)
{
	switch (c) {
#ifdef VI_KEYS /* hehe. Not done yet, though */
	case 'h':			return CTRL('b');
	case 'j':			return CTRL('n');
	case 'k':			return CTRL('p');
	case 'l':			return CTRL('f');
	case '0':			return CTRL('a');
	case '$':			return CTRL('e');
	case '/':			return CTRL('s');
	case '?':			return CTRL('r');
#endif
#ifdef KEY_MIN
	case KEY_BEG: case KEY_FIND:
	case KEY_HOME:			return CTRL('a');
	case KEY_LEFT:			return CTRL('b');
	case KEY_END: case KEY_SELECT:	return CTRL('e');
	case KEY_RIGHT:			return CTRL('f');
	case KEY_EOL:			return CTRL('k');
	case KEY_DC:			return CTRL('d');
	case KEY_DOWN:			return CTRL('n');
	case KEY_UP:			return CTRL('p');
	case KEY_PPAGE:			return ALT('v');
	case KEY_NPAGE:			return CTRL('v');
	case KEY_SUSPEND:		return CTRL('z');
	case KEY_BACKSPACE: 		return CTRL('h');
	case KEY_CANCEL:		return CTRL('g');
	case KEY_ENTER: 		return '\n';
#endif
	case '\b':			return CTRL('h');
	case '\r':			return CTRL('\n');
	default:			return c;
	}
	/*NOTREACHED*/
}

/*X
Reads and returns one input character.
If there are characters in the input queue, these are used first.
If macro_flag is TRUE, the character is recorded.
X*/
int get_char_from_input_queue()
{
	int c;
  
	if (input_queue.size > 0) {
		c = input_queue.buf[input_queue.index++];
		input_queue.index %= IQ_MAX;
		input_queue.size--;
	} else {
		c = fix_input(wgetch(single.win));
	}
	if (macro_flag) record_macro(c);

	TRACEME((f, "\tgcfiq: %d", c));

	return c;
}

void remember_point(window *w)
{
	oldr = get_point(w).row;
	oldc = get_point(w).col;
}

void mainloop()
{
	TRACEME((f, "Now entering the main loop!"));

	remember_point(w_list);
	pr_scr_flag = TRUE;

	for(;;) {
		show_cur(w_list);
		remember_point(w_list);
		do_cmd(get_char_from_input_queue());
	}
	/*NOTREACHED*/
}

/* Prints the string p on the bottom line of the screen. */
void llpr(char *p) /* important */
{
	TRACEME((f, "this is llpr speaking (%s)", p));

	wmove(single.win, single.size.ws_row - 1, 0);
	waddnstr(single.win, p, single.size.ws_col);
	wclrtoeol(single.win);
	wrefresh(single.win);
}

/* Remove w from the window list and free the memory it used.
If w was the active window, w_list is moved forward to the next window.
If w was the last window, w_list is set to NULL. */
void free_window(window *w)
{
	window *pw;
  
	TRACEME((f, "free_window"));

	/* unlink from window list */
	for (pw = w_list; next_window(pw) != w && next_window(pw) != pw; pw = next_window(pw));

	next_window(pw) = next_window(w);
  
	/* make sure w_lst does not point to a deleted window */
	if (w_list == w) w_list = next_window(w_list);

	/* no windows in the list => w_list = NULL */
	if (w_list == w) w_list = NULL;

	cfree(w);
}

/*X
The new window points to buffer b.  Point and top positions are set
to the first position in b.
The new window is inserted in the window list after prev, or after itself
if prev is NULL.
The new window is returned, or NULL, if it couldn't be created.
X*/
window *new_window(buffer *b, window *prev)
{
	window *w;

	TRACEME((f, "new_window"));

	if((w = (window *)cmalloc(sizeof(window))) == NULL) return NULL;

	set_window_buffer(w, b);
	w->sht = w->bsht = 0;

	if (prev == NULL) prev = w;
	else next_window(w) = next_window(prev);

	next_window(prev) = w; /* macros are fun! */

	set_point_row(w, 1); set_point_col(w, 1);
	set_blku_row(w, 2);  set_blku_col(w, 2);
	set_blkl_row(w, 1);  set_blkl_col(w, 1);

	set_top_row(w, 1); /* use same value as point_row above */
	set_top_col(w, 1); /* use same value as point_col above */

	set_prot_row(w, 1); /* again, use same values */
	set_prot_col(w, 1); /* again, use same values */

	return w;
}

void die(char *s)
{
	TRACEME((f, "Die: %s", s));
	exit_windows();			/* Reset the terminal, */
	fprintf(stderr, "%s\n", s);	/* tell'm what happened, */
	exit(EXIT_FAILURE);		/* and split! */
}

/* Internal */
void resize_ncurses()
{
	struct winsize size;

	TRACEME((f, "resize_ncurses"));
#ifdef HAVE_RESIZETERM
	if (ioctl(fileno(stdout), TIOCGWINSZ, &size) == 0 &&
	    resizeterm(size.ws_row, size.ws_col) == OK) {

		/* Since everything is full screen right now, this works */
		single.size.ws_row = size.ws_row;
		single.size.ws_col = size.ws_col;
	} else {
#endif
		/* Not so cool if this happens */
		single.size.ws_row = LINES;
		single.size.ws_col = COLS;
#ifdef HAVE_RESIZETERM
	}
#endif

	/* The code doesn't handle small screens very well, so for now we
	   just bail out right away.  These dimentions hopefully aren't
	   too limiting */
	if (single.size.ws_col < 30 || single.size.ws_row < 10) 
		die("The screen is too small");

	width = 8;
	left_margin = 8;
	columns = (single.size.ws_col - left_margin) / width;

	pr_scr_flag = TRUE;
	show_cur(w_list);
} /* resize_ncurses */

void new_ncurses()
{
	TRACEME((f,"new_ncurses"));

	/* the four 0's mean make it full screan */
	if((single.win = newwin(0, 0, 0, 0)) == NULL)
		die("Error activating window");
#ifdef HAVE_IMMEDOK
	immedok(single.win, FALSE);
#endif
	scrollok(single.win, FALSE);
	leaveok(single.win, FALSE);
#ifdef HAVE_KEYPAD
	keypad(single.win, TRUE);
#endif
	resize_ncurses();
}

/* Much nicer then attempting to spawn a shell :) */
LISP suspend_function()
{
	TRACEME((f,"suspend_function"));
#ifdef SIGTSTP
	delwin(single.win);
	raise(SIGTSTP);
	new_ncurses();
#endif
	return NIL;
}

/* What to do when someone resizes the xterm (for example) */
void sig_winch_handler()
{
	TRACEME((f,"sig_winch_handler"));
	wclear(single.win);
	resize_ncurses();
}

/* What to do when someone does a 'killall tsiag' (for example) */
/* Eventually, we'll want to save data somewhere where it can be recovered */
void sig_term_handler()
{
	die("Recieved SIGTERM");
}

/* Needed for scroll-up/down/left/right */
LISP get_geometry(void)
{
	int x, y;
	unsigned int width, height, border_width, depth;
	LISP result;

	TRACEME((f, "get_geometry"));

	/* Say the magic words... */
	x = y = border_width = depth = 0;
	width = (single.size.ws_col - 3) * 10;
	height = (single.size.ws_row - 3) * 20;

	result = cons(flocons(depth), NIL);
	result = cons(flocons(border_width), result);
	result = cons(flocons(height), result);
	result = cons(flocons(width), result);
	result = cons(flocons(y), result);
	result = cons(flocons(x), result);

	return result;
}

/* Sets up the whole initial window structure and initializes scrupd.
The window list w_list is set to a list with a single window with the
buffer b. */
void init_windows(buffer *b, int *argc, char **argv)
{
	TRACEME((f,"init_windows"));

	initscr();    /* this curses function must be called first */

	raw();        /* Make input behave */
	noecho();
	nonl();

	/* very important */
	interp_startup();
	init_subr_0("suspend-function", suspend_function);
	init_subr_0("get-geometry", get_geometry);

#if 0
	init_form();
#endif

#ifdef SIGWINCH
	signal(SIGWINCH, sig_winch_handler);
#endif
#ifdef SIGTERM
	signal(SIGTERM, sig_term_handler);
#endif

	w_list = new_window(b, (window *)0);

	new_ncurses();

	init_calc_cmds();

	encode_format(~0, &fmt0);
	rowcol_fmt_index = encode_format(~0, &fmt1);

	kbd_macro.maxsize = KB_MAX;
	kbd_macro.text = (char *)ccalloc(KB_MAX, sizeof(char));
  
	input_queue.buf = (char *)ccalloc(IQ_MAX, sizeof(char));
	input_queue.index = 0;
	input_queue.size = 0;
}

/* Cleans up after Calc before exit.  All buffers and windows are freed. */
void exit_windows()
{
	TRACEME((f, "exit_windows"));

	/* free all buffers */
	while (b_list != NULL) free_buffer(b_list);
  
	/* free all windows */
	while (w_list != NULL) free_window(w_list);

	endwin();
}

/* This is significantly different then the xsiag draw_input */
/* Also, it's duplicating some of what siag/cmds.c does */
/* but it shure do look pretty */
void draw_input(char *b, int n)
{
	int s = w_list->sht;
	int r = get_point(w_list).row;
	int c = get_point(w_list).col;

	switch (ret_type(buffer_of_window(w_list), s, r, c)) {
	case LABEL: 
		sprintf(b, "Label:");
		strncat(b, ret_text(buffer_of_window(w_list), s, r, c), n - 6);
		break;
	case ERROR: case EXPRESSION: case STRING:
		sprintf(b, "%s expression:", interpreter2name(ret_interpreter(buffer_of_window(w_list), s, r, c)));
		strncat(b, ret_text(buffer_of_window(w_list), s, r, c), n - 14);
		break;
	case EMPTY: b[0] = '\0'; return;
	case EMBED: sprintf(b, "Embedded object"); return;
	default: sprintf(b, "Unkown object type"); return;
	}
	return;
}

/* Move the cursor to reflect the position of point in w.
If point is not visible, the window is moved so that point is in
the middle of the screen. */
void show_cur(window *w)
{
	char b[single.size.ws_col]; /* Is there a better way to do this? */
	int i;

	TRACEME((f,"\tshow_cur(): enter: point=(%d,%d) top=(%d,%d)",
	         get_point(w).row, get_point(w).col,
	         get_top(w).row, get_top(w).col));

	if (get_point(w).row < get_top(w).row || get_point(w).row > get_top(w).row + (single.size.ws_row - 4)) {
		set_top_row(w, get_point(w).row - ((single.size.ws_row - 5) / 2));
		pr_scr_flag = TRUE;
	}

	if (get_top(w).row < 1) {
		set_top_row(w, 1);
		pr_scr_flag = TRUE;
	} else if (get_top(w).row > BUFFER_ROWS - (single.size.ws_row - 5)) {
		set_top_row(w, BUFFER_ROWS - (single.size.ws_row - 4));
		pr_scr_flag = TRUE;
	}

	if (get_point(w).col < get_top(w).col ||
	    get_point(w).col > get_top(w).col + (columns - 1)) {
		set_top_col(w, get_point(w).col - (columns / 2));
		pr_scr_flag = TRUE;
	}

	if (get_top(w).col < 1) {
		set_top_col(w, 1);
		pr_scr_flag = TRUE;
	} else if (get_top(w).col > BUFFER_COLS - (columns - 1)) {
		set_top_col(w, BUFFER_COLS - (columns - 1));
		pr_scr_flag = TRUE;
	}

	TRACEME((f,"\tshow_cur() point=(%d,%d) top=(%d,%d) pr_scr_flag=(%d)",
	         get_point(w).row, get_point(w).col,
	         get_top(w).row, get_top(w).col, pr_scr_flag));

	if (pr_scr_flag) pr_scr(w);

	/* Print contents of cell */
	if(get_point(w_list).row != oldr || get_point(w_list).col != oldc) {
		draw_input(b, single.size.ws_col);
		wmove(single.win, single.size.ws_row - 1, 0);
		waddnstr(single.win, b, single.size.ws_col);
		wclrtoeol(single.win);
	}

	wstandout(single.win);

	/* the line */
	wmove(single.win, single.size.ws_row - 2, 0);
	for (i = 0; i < single.size.ws_col; i++)
		waddch(single.win, '-');

	/* filename and modified flag */
	wmove(single.win, single.size.ws_row - 2, 0);
	wprintw(single.win, "[%s]%s%s",
	        buffer_of_window(w)->name,
	        buffer_of_window(w)->change ? "[M]" : "",
	        macro_flag ? "[Record Macro]" : "");

	/* cursor coordinates */
	sprintf(b, "[%d, %d]", get_point(w).row, get_point(w).col);
	wmove(single.win, single.size.ws_row - 2, single.size.ws_col - strlen(b));
	waddstr(single.win, b);

	wstandend(single.win);

	/* finally, the cursor */
	wmove(single.win, 1 + get_point(w).row - get_top(w).row,
	      left_margin + width * (get_point(w).col - get_top(w).col));

	wrefresh(single.win);
} /* show_cur */

/* This function is used */
void activate_window(window *w)
{
	TRACEME((f,"activate_window"))
	w_list = w;
}

int select_from_list(char *text, char **choices, int nchoices)
{
	int i;

	TRACEME((f,"select_from_list(%s,%s...,%d)",text,*choices,nchoices))

	wclear(single.win);

	wmove(single.win, 0, 0);
	wstandout(single.win);
	waddnstr(single.win, text, single.size.ws_col);
	for (i = strlen(text); i < single.size.ws_col; i++) waddch(single.win, ' ');
	wstandend(single.win);

	for (i = 0; i < nchoices; i++) {
		wmove(single.win, i + 2, 2);
		wprintw(single.win, "%d: %s", i, choices[i]);
	}

	wmove(single.win, nchoices + 3, 0);

	wrefresh(single.win);
	pr_scr_flag = TRUE;

	for(;;) {
		if ((i = get_char_from_input_queue()) == ESC) return -1;

		if (i >= '0' && i < nchoices + '0') { 
			wmove(single.win, i - '0' + 2, 2);
			wstandout(single.win);
			wprintw(single.win, "%d: %s", i - '0', choices[i - '0']);
			wstandend(single.win);
			return i - '0';
		}
	}
	/*NOTREACHED*/
}

